#!/usr/bin/perl

package Cognition::Export::Contact;
use Cognition::Export;
use Cognition::Export::Address;
use Cognition::Export::Point;
use HTML::Entities qw(encode_entities_numeric);
use MIME::Base64;
use URI;

# Reference:
# http://www.ibiblio.org/hhalpin/homepage/notes/vcardtable.html
# (Though that page does contain a number of mistakes, such as equating a vCard
# "TITLE" [a job title] with a FOAF "title" [e.g. Mr, Mrs].)

sub is
{
	my $subject = shift;
	return unless ($subject);
	return
		unless (defined $subject->property('http://www.w3.org/1999/02/22-rdf-syntax-ns#type'));
	foreach my $Type ( @{$subject
		->property('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')
		->valObjects} )
	{
		my $type = $Type->toString;
		return 1
			if (($type eq 'http://xmlns.com/foaf/0.1/Person')
			||  ($type eq 'http://xmlns.com/foaf/0.1/Organisation')
			||  ($type eq 'http://xmlns.com/foaf/0.1/Project')
			||  ($type eq 'http://xmlns.com/foaf/0.1/Group')
			||  ($type eq 'http://www.w3.org/2001/vcard-rdf/3.0#VCARD')
			||  ($type eq 'http://www.w3.org/2006/vcard/ns#VCard')
			||	 ($type eq 'http://www.w3.org/2006/03/hcard#vcard')
			||  ($type eq 'http://www.w3.org/2000/10/swap/pim/contact#Person')
			||  ($type eq 'http://www.w3.org/2000/10/swap/pim/contact#Male')
			||  ($type eq 'http://www.w3.org/2000/10/swap/pim/contact#Female')
			||  ($type eq 'http://www.w3.org/2000/10/swap/pim/contact#SocialEntity')
			||  ($type eq 'urn:ietf:rfc:2426#vcard')
			||  ($type eq 'http://xmlns.com/wot/0.1/#term_User'));
	}
	return;
}

sub outputs
{
	return [
		{
			code  => 'VCF',
			name  => 'vCard',
			type  => 'text/directory',
			multi => 1
		},
		{
			code  => 'Name',
			name  => 'Contact Name',
			type  => 'text/plain',
			multi => 1
		}
	];
}

sub multiVCF
{
	my $triples = shift;
	my $rv = '';
	foreach my $s (keys %$triples)
	{
		$rv .= outVCF($triples, $s)
			if (defined $triples->{$s} && is($triples->{$s}));
	}
	return $rv;
}

sub outVCF
{
	my $triples = shift;
	my $subject = shift;
	my $S = $triples->{$subject};
	my @rv = (['BEGIN','VCARD']);

	# UID
	my $uid = firstOf($S, 'urn:ietf:rfc:2426#uid',
		'http://www.w3.org/2006/vcard/ns#uid',
		'http://www.w3.org/2006/03/hcard#uid',
		'http://www.w3.org/2001/vcard-rdf/3.0#uid',
		'http://purl.org/dc/terms/identifier',
		'http://purl.org/dc/elements/1.1/identifier');
	push @rv, ['UID', $uid]
		if ($uid);
	
	# Name components
	my $honorific_pfx = firstOf($S, 'urn:ietf:rfc:2426#honorific-prefix',
		'http://www.w3.org/2006/vcard/ns#honorific-prefix',
		'http://www.w3.org/2006/03/hcard#honofific-prefix',
		'http://www.w3.org/2001/vcard-rdf/3.0#honorific-prefix',
		'http://xmlns.com/foaf/0.1/title',
		'http://www.w3.org/2000/10/swap/pim/contact#personalTitle');
	my $firstname = firstOf($S, 'urn:ietf:rfc:2426#given-name',
		'http://www.w3.org/2006/vcard/ns#given-name',
		'http://www.w3.org/2006/03/hcard#given-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#given-name',
		'http://xmlns.com/foaf/0.1/givenname',
		'http://www.w3.org/2000/10/swap/pim/contact#firstName');
	my $middlename = firstOf($S, 'urn:ietf:rfc:2426#additional-name',
		'http://www.w3.org/2006/vcard/ns#additional-name',
		'http://www.w3.org/2006/03/hcard#additional-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#additional-name');
	my $surname = firstOf($S, 'urn:ietf:rfc:2426#family-name',
		'http://www.w3.org/2006/vcard/ns#family-name',
		'http://www.w3.org/2006/03/hcard#family-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#family-name',
		'http://xmlns.com/foaf/0.1/family_name',
		'http://xmlns.com/foaf/0.1/surname',
		'http://www.w3.org/2000/10/swap/pim/contact#lastName');
	my $honorific_sfx = firstOf($S, 'urn:ietf:rfc:2426#honorific-suffix',
		'http://www.w3.org/2006/vcard/ns#honorific-suffix',
		'http://www.w3.org/2006/03/hcard#honorific-suffix',
		'http://www.w3.org/2001/vcard-rdf/3.0#honorific-suffix',
		'http://www.w3.org/2000/10/swap/pim/contact#personalSuffix');
		
	# Formatted name
	my $fn = firstOf($S, 'urn:ietf:rfc:2426#fn',
		'http://www.w3.org/2006/vcard/ns#fn',
		'http://www.w3.org/2006/03/hcard#fn',
		'http://www.w3.org/2001/vcard-rdf/3.0#fn',
		'http://xmlns.com/foaf/0.1/fullName',
		'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
		'http://purl.org/dc/terms/title',
		'http://purl.org/dc/elements/1.1/title');	
	if (!$fn)
		{ $fn = "$firstname $surname"; }
	push @rv, ['FN',$fn];

	# Structured name
	my $n = firstOf($S, 'urn:ietf:rfc:2426#n',
		'http://www.w3.org/2006/vcard/ns#n',
		'http://www.w3.org/2006/03/hcard#n',
		'http://www.w3.org/2001/vcard-rdf/3.0#n');	
	if (!$n)
	{
		$n = "$surname;$firstname;$middlename;$honorific_pfx;$honorific_sfx";
		$n =~ s/ /\,/;
	}
	push @rv, ['N',$n,1];
	
	# Nickname
	my @nicknames = allOf($S, 'urn:ietf:rfc:2426#nickname',
		'http://www.w3.org/2006/vcard/ns#nickname',
		'http://www.w3.org/2006/03/hcard#nickname',
		'http://www.w3.org/2001/vcard-rdf/3.0#nickname',
		'http://xmlns.com/foaf/0.1/nick',
		'http://www.w3.org/2000/10/swap/pim/contact#knownAs');
	foreach my $nickname (@nicknames)
		{ push @rv, ['NICKNAME',$nickname]; }

	# Sort-String
	my $sortstring = firstOf($S, 'urn:ietf:rfc:2426#sort-string',
		'http://www.w3.org/2006/vcard/ns#sort-string',
		'http://www.w3.org/2006/03/hcard#sort-string',
		'http://www.w3.org/2001/vcard-rdf/3.0#sort-string',
		'http://www.w3.org/2000/10/swap/pim/contact#sortName');
	push @rv, ['SORT-STRING',$sortstring]
		if ($sortstring);

	# Title
	my @titles = allOf($S, 'urn:ietf:rfc:2426#title',
		'http://www.w3.org/2006/vcard/ns#title',
		'http://www.w3.org/2006/03/hcard#title',
		'http://www.w3.org/2001/vcard-rdf/3.0#title');
	foreach my $title (@titles)
		{ push @rv, ['TITLE',$title]; }

	# Role
	my @roles = allOf($S, 'urn:ietf:rfc:2426#role',
		'http://www.w3.org/2006/vcard/ns#role',
		'http://www.w3.org/2006/03/hcard#role',
		'http://www.w3.org/2001/vcard-rdf/3.0#role');
	foreach my $role (@roles)
		{ push @rv, ['ROLE',$role]; }

	# Addresses
	my @adrs = allOfObj($S, 'urn:ietf:rfc:2426#adr',
		'http://www.w3.org/2006/vcard/ns#adr',
		'http://www.w3.org/2006/vcard/ns#homeAdr',
		'http://www.w3.org/2006/vcard/ns#workAdr',
		'http://www.w3.org/2006/vcard/ns#unlabelledAdr',
		'http://www.w3.org/2006/03/hcard#adr',
		'http://www.w3.org/2001/vcard-rdf/3.0#adr',
		'http://www.w3.org/2000/10/swap/pim/contact#address');
	foreach my $adr (@adrs)
	{
		if ($adr->type && $triples->{$adr->toString})
		{
			my $adrSubject = $triples->{$adr->toString};
			if (Cognition::Export::Address::is($adrSubject))
			{
				my ($key, $val) = Cognition::Export::Address::outVCFLine($triples, $adr->toString);
				push @rv, [$key, $val, 1];
			}
		}
		else
		{
			my $a = $adr->toString;
			$a =~ s/(^\s+|\s+$)//g;
			$a =~ s/\s+/ /g;
			push @rv, ['LABEL',$a] if ($a);
		}
	}

	# Geo
	my @geos = allOfObj($S, 'urn:ietf:rfc:2426#geo',
		'http://www.w3.org/2006/vcard/ns#geo',
		'http://www.w3.org/2006/03/hcard#geo',
		'http://www.w3.org/2001/vcard-rdf/3.0#geo');
	foreach my $geo (@geos)
	{
		if ($geo->type && $triples->{$geo->toString})
		{
			my $geoSubject = $triples->{$geo->toString};
			if (Cognition::Export::Point::is($geoSubject))
			{
				my ($key, $val) = Cognition::Export::Point::outVCFLine($triples, $geo->toString);
				push @rv, [$key, $val, 1];
			}
		}
	}

	# Labels
	my @labels = allOf($S, 'urn:ietf:rfc:2426#label',
		'http://www.w3.org/2006/vcard/ns#label',
		'http://www.w3.org/2006/03/hcard#label',
		'http://www.w3.org/2001/vcard-rdf/3.0#label');
	foreach my $label (@labels)
		{ push @rv, ['LABEL',$label]; }

	# Telephone numbers
	my @tels = allOfObj($S, 'urn:ietf:rfc:2426#tel',
		'http://www.w3.org/2006/vcard/ns#tel',
		'http://www.w3.org/2006/vcard/ns#homeTel',
		'http://www.w3.org/2006/vcard/ns#workTel',
		'http://www.w3.org/2006/vcard/ns#mobileTel',
		'http://www.w3.org/2006/vcard/ns#unlabelledTel',
		'http://www.w3.org/2006/vcard/ns#fax',
		'http://www.w3.org/2006/03/hcard#tel',
		'http://www.w3.org/2001/vcard-rdf/3.0#tel',
		'http://www.w3.org/2000/10/swap/pim/contact#phone',
		'http://www.w3.org/2000/10/swap/pim/contact#mobile',
		'http://xmlns.com/foaf/0.1/phone');
	foreach my $tel (@tels)
	{
		my $telSubject;
		if ($tel->type && $triples->{$tel->toString})
		{
			$telSubject = $triples->{$tel->toString};
		}

		my $usages;
		$usages = 'CELL'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#mobileTel');
		$usages = 'CELL'
			if ($tel->property->uri eq 'http://www.w3.org/2000/10/swap/pim/contact#mobile');
		$usages = 'FAX'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#fax');
		$usages = 'HOME'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#homeTel');
		$usages = 'WORK'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#workTel');
			
		if ($telSubject)
		{
			my @usages = allOf($telSubject, 'urn:ietf:rfc:2426#type');
			$usages = uc(join ',', @usages);
		}
		
		my $telno = $tel->toString;
		if ($telno =~ /^tel\:/)
		{
			$telno =~ s/^tel\://;
			$telno =~ s/\;.*$//g;
		}
		
		if ($usages)
			{ push @rv, ["TEL;TYPE=$usages",$telno]; }
		else
			{ push @rv, ['TEL', $telno]; }
	}

	# E-mail addresses. W3C vocab defines a bunch of weird e-mail types which
	# have nothing to do with vCard. This procedure is almost identical to "TEL:"
	# but I don't think we can wrap it all in a function.
	my @emails = allOfObj($S, 'urn:ietf:rfc:2426#email',
		'http://www.w3.org/2006/vcard/ns#email',
		'http://www.w3.org/2006/vcard/ns#workEmail',
		'http://www.w3.org/2006/vcard/ns#mobileEmail',
		'http://www.w3.org/2006/vcard/ns#unlabelledEmail',
		'http://www.w3.org/2006/03/hcard#email',
		'http://www.w3.org/2001/vcard-rdf/3.0#email',
		'http://www.w3.org/2000/10/swap/pim/contact#emailAddress',
		'http://xmlns.com/foaf/0.1/mbox',
		'http://xmlns.com/foaf/0.1/mbox_sha1sum');
	foreach my $email (@emails)
	{
		my $emailSubject;
		if ($email->type && $triples->{$email->toString})
		{
			$emailSubject = $triples->{$email->toString};
		}

		my $usages;
		if ($emailSubject)
		{
			my @usages = allOf($emailSubject, 'urn:ietf:rfc:2426#type');
			$usages = uc(join ',', @usages);
		}
		
		my $emailno = $email->toString;
		$usages = 'X-SHA1SUM'
			if ($email->property->uri eq 'http://xmlns.com/foaf/0.1/mbox_sha1sum');
		if ($emailno =~ /^mailto\:/)
		{
			$emailno =~ s/^mailto\://;
			$emailno =~ s/\?.*$//g;
		}
		
		if ($usages)
			{ push @rv, ["EMAIL;TYPE=$usages",$emailno]; }
		else
			{ push @rv, ['EMAIL',$emailno]; }
	}

	# URLs
	my @urls = allOf($S, 'urn:ietf:rfc:2426#url',
		'http://www.w3.org/2006/vcard/ns#url',
		'http://www.w3.org/2006/03/hcard#url',
		'http://www.w3.org/2001/vcard-rdf/3.0#url',
		'http://xmlns.com/foaf/0.1/homepage',
		'http://xmlns.com/foaf/0.1/weblog',
		'http://xmlns.com/foaf/0.1/workplaceHomepage',
		'http://xmlns.com/foaf/0.1/workInfoHomepage',
		'http://xmlns.com/foaf/0.1/schoolHomepage',
		'http://xmlns.com/foaf/0.1/tipjar',
		'http://www.w3.org/2000/10/swap/pim/contact#homepage',
		'http://www.w3.org/2000/10/swap/pim/contact#publicHomepage');
	foreach my $url (@urls)
		{ push @rv, ['URL', $url]; }

	# Org
	my @orgs = allOfObj($S, 'urn:ietf:rfc:2426#org',
		'http://www.w3.org/2006/vcard/ns#org',
		'http://www.w3.org/2006/03/hcard#org',
		'http://www.w3.org/2001/vcard-rdf/3.0#org');
	foreach my $org (@orgs)
	{
		my ($orgSubject, $orgstring);
		
		if ($org->type && $triples->{$org->toString})
		{
			$orgSubject = $triples->{$org->toString};

			if ($orgSubject)
			{
				$orgstring = firstOf($orgSubject,
					'urn:ietf:rfc:2426#organization-name',
					'http://www.w3.org/2006/vcard/ns#organization-name',
					'http://www.w3.org/2006/03/hcard#organization-name',
					'http://www.w3.org/2001/vcard-rdf/3.0#organization-name');
				$orgstring =~ s/\;/\\\;/g;
				$orgstring =~ s/\,/\\\,/g;

				my @units = allOf($orgSubject,
					'urn:ietf:rfc:2426#organization-unit',
					'http://www.w3.org/2006/vcard/ns#organization-unit',
					'http://www.w3.org/2006/03/hcard#organization-unit',
					'http://www.w3.org/2001/vcard-rdf/3.0#organization-unit');
				foreach my $ou (@units)
				{
					$ou =~ s/\;/\\\;/g;
					$ou =~ s/\,/\\\,/g;
					$orgstring .= "; $ou";
				}
			}
		}
		
		$orgstring = $org->toString
			unless (length $orgstring);

		push @rv, ['ORG', $orgstring, 1] if ($orgstring);
	}

	# Mailer
	my @mailers = allOf($S, 'urn:ietf:rfc:2426#mailer',
		'http://www.w3.org/2006/vcard/ns#mailer',
		'http://www.w3.org/2006/03/hcard#mailer',
		'http://www.w3.org/2001/vcard-rdf/3.0#mailer');
	foreach my $mailer (@mailers)
		{ push @rv, ['MAILER',$mailer]; }

	# Photo
	my @photos = allOf($S, 'urn:ietf:rfc:2426#photo',
		'http://www.w3.org/2006/vcard/ns#photo',
		'http://www.w3.org/2006/03/hcard#photo',
		'http://www.w3.org/2001/vcard-rdf/3.0#photo',
		'http://xmlns.com/foaf/0.1/img',
		'http://xmlns.com/foaf/0.1/depiction');
	foreach my $photo (@photos)
		{ push @rv, ['PHOTO;VALUE=uri',$photo]; }

	# Logo
	my @logos = allOf($S, 'urn:ietf:rfc:2426#logo',
		'http://www.w3.org/2006/vcard/ns#logo',
		'http://www.w3.org/2006/03/hcard#logo',
		'http://www.w3.org/2001/vcard-rdf/3.0#logo',
		'http://xmlns.com/foaf/0.1/logo');
	foreach my $logo (@logos)
		{ push @rv, ['LOGO;VALUE=uri',$logo]; }

	# Sound
	my @sounds = allOf($S, 'urn:ietf:rfc:2426#sound',
		'http://www.w3.org/2006/vcard/ns#sound',
		'http://www.w3.org/2006/03/hcard#sound',
		'http://www.w3.org/2001/vcard-rdf/3.0#sound');
	foreach my $sound (@sounds)
		{ push @rv, ['SOUND;VALUE=uri',$sound]; }

	# Key (crypto)
	my @keys = allOf($S, 'urn:ietf:rfc:2426#key',
		'http://www.w3.org/2006/vcard/ns#key',
		'http://www.w3.org/2006/03/hcard#key',
		'http://www.w3.org/2001/vcard-rdf/3.0#key');
	foreach my $key (@keys)
		{ push @rv, ['KEY;VALUE=uri',$key]; }
	
	# Keys (WOT RDF schema)
	my @keys = allOfObj($S, 'http://xmlns.com/wot/0.1/hasKey');
	foreach my $key (@keys)
	{
		if ($key->type && defined $triples->{$key->toString})
		{
			my $uri = firstOf($triples->{$key->toString},
				'http://xmlns.com/wot/0.1/pubkeyAddress');
			push @rv, ['KEY;VALUE=uri', $uri]
				if ($uri);
		}
	}

	# Notes
	my @notes = allOf($S, 'urn:ietf:rfc:2426#note',
		'http://www.w3.org/2006/vcard/ns#note',
		'http://www.w3.org/2006/03/hcard#note',
		'http://www.w3.org/2001/vcard-rdf/3.0#note',
		'http://www.w3.org/2000/01/rdf-schema#comment',
		'http://purl.org/dc/terms/abstract',
		'http://purl.org/dc/terms/description',
		'http://purl.org/dc/elements/1.1/description',
		'http://xmlns.com/foaf/0.1/plan',
		'http://xmlns.com/foaf/0.1/geekcode');
	foreach my $note (@notes)
		{ push @rv, ['NOTE',$note]; }

	# Agents
	my @agents = allOfObj($S, 'urn:ietf:rfc:2426#agent');
	foreach my $agent (@agents)
	{
		if ($agent->type && $triples->{$agent->toString})
		{
			my $agentSubject = $triples->{$agent->toString};
			if (is($agentSubject))
			{
				my $a = outVCF($triples, $agent->toString);
				push @rv, ['AGENT',$a];
			}
			else
			{
				my $a = firstOf($agentSubject,
					'urn:ietf:rfc:2426#fn',
					'http://www.w3.org/2006/vcard/ns#fn',
					'http://www.w3.org/2006/03/hcard#fn',
					'http://www.w3.org/2001/vcard-rdf/3.0#fn',
					'http://xmlns.com/foaf/0.1/fullName',
					'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
					'http://purl.org/dc/terms/title',
					'http://purl.org/dc/elements/1.1/title');
				push @rv, ['AGENT;VALUE=text',$a] if ($a);
			}
		}
		else
		{
			my $a = $agent->toString;
			$a =~ s/(^\s+|\s+$)//g;
			$a =~ s/\s+/ /g;
			push @rv, ['AGENT;VALUE=text',$a] if ($a);
		}
	}

	# TZ (timezone)
	my $tz = firstOf($S, 'urn:ietf:rfc:2426#tz',
		'http://www.w3.org/2006/vcard/ns#tz',
		'http://www.w3.org/2006/03/hcard#tz',
		'http://www.w3.org/2001/vcard-rdf/3.0#tz');
	push @rv, ['TZ',$tz]
		if ($tz);

	# Bday (date of birth)
	my $bday = firstOf($S, 'urn:ietf:rfc:2426#bday',
		'http://www.w3.org/2006/vcard/ns#bday',
		'http://www.w3.org/2006/03/hcard#bday',
		'http://www.w3.org/2001/vcard-rdf/3.0#bday',
		'http://www.w3.org/2000/10/swap/pim/contact#birthday',
		'http://xmlns.com/foaf/0.1/birthday');
	push @rv, ['BDAY',$bday]
		if ($bday);
		
	# Gender
	my $gender = firstOf($S, 'urn:ietf:rfc:2426#x-gender',
		'http://www.w3.org/2006/vcard/ns#x-gender',
		'http://www.w3.org/2006/03/hcard#x-gender',
		'http://www.w3.org/2001/vcard-rdf/3.0#x-gender',
		'http://xmlns.com/foaf/0.1/gender');
	if ($gender)
	{
		push @rv, ['X-GENDER',$gender];
	}
	else
	{
		if (inProp($S, 'http://www.w3.org/2000/10/swap/pim/contact#Male',
			'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'))
		{
			push @rv, ['X-GENDER','Male'];
		}

		if (inProp($S, 'http://www.w3.org/2000/10/swap/pim/contact#Female',
			'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'))
		{
			push @rv, ['X-GENDER','Female'];
		}
	}
	
	# Categories
	my @cats = allOfObj($S, 'urn:ietf:rfc:2426#category',
		'http://www.w3.org/2006/vcard/ns#category',
		'http://www.w3.org/2006/03/hcard#category',
		'http://www.w3.org/2001/vcard-rdf/3.0#category');
	my $catstring = '';
	foreach my $cat (@cats)
	{
		if ($cat->type && $triples->{$cat->toString})
		{
			my $catSubject = $triples->{$cat->toString};
			my $category = firstOf($catSubject,
				'http://microformats.org/wiki/rel-tag#tag',
				'http://purl.org/dc/terms/description',
				'http://purl.org/dc/elements/1.1/description');
			$catstring .= ','.uc($category)
				if (length $category);
		}
		else
		{
			my $category = $cat->toString;
			$category =~ s/(^\s+|\s+$)//g;
			$category =~ s/\s+/ /g;
			$catstring .= ','.uc($category);
		}
	}
	$catstring =~ s/^\,//;
	push @rv, ['CATEGORIES',$catstring,1]
		if ($catstring);

	# Class
	my $class = uc(firstOf($S, 'urn:ietf:rfc:2426#class',
		'http://www.w3.org/2006/vcard/ns#class',
		'http://www.w3.org/2006/03/hcard#class',
		'http://www.w3.org/2001/vcard-rdf/3.0#class'));
	push @rv, ['CLASS',$class]
		if ($class =~ /^(PUBLIC|PRIVATE|CONFIDENTIAL|X-)/);
		
	# Rev
	my $rev = uc(firstOf($S, 'urn:ietf:rfc:2426#rev',
		'http://www.w3.org/2006/vcard/ns#rev',
		'http://www.w3.org/2006/03/hcard#rev',
		'http://www.w3.org/2001/vcard-rdf/3.0#rev'));
	push @rv, ['REV',$rev]
		if ($rev);
		
	# Done.
	push @rv, ['END','VCARD'];
	
	# Convert to text/directory syntax.
	my $rv = '';
	foreach my $line (@rv)
	{
		my ($k, $v, $soft) = @{$line};
		
		if ($k =~ /VALUE.uri$/i && $v =~ /^data\:/)
		{
			my $uri = URI->new($v);
			$v = encode_base64($uri->data, '');
			$k =~ s/VALUE.uri$/ENCODING\=b/i;
		}
		
		$v =~ s/\\/\\\\/g;
		$v =~ s/\r?\n/\\n/g;
		unless ($soft)
		{
			$v =~ s/\;/\\\;/g;
			$v =~ s/\,/\\\,/g;
		}
		
		$rv .= "$k:$v\r\n";
	}
	return $rv;
}


sub outName
{
	my $triples = shift;
	my $subject = shift;
	my $S = $triples->{$subject};
	
	my $fn = firstOf($S, 'urn:ietf:rfc:2426#fn',
		'http://www.w3.org/2006/vcard/ns#fn',
		'http://www.w3.org/2006/03/hcard#fn',
		'http://www.w3.org/2001/vcard-rdf/3.0#fn',
		'http://xmlns.com/foaf/0.1/fullName',
		'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
		'http://purl.org/dc/terms/title',
		'http://purl.org/dc/elements/1.1/title');	
	if (length $fn)
		{ return $fn; }

	# Name components
	my $honorific_pfx = firstOf($S, 'urn:ietf:rfc:2426#honorific-prefix',
		'http://www.w3.org/2006/vcard/ns#honorific-prefix',
		'http://www.w3.org/2006/03/hcard#honofific-prefix',
		'http://www.w3.org/2001/vcard-rdf/3.0#honorific-prefix',
		'http://xmlns.com/foaf/0.1/title',
		'http://www.w3.org/2000/10/swap/pim/contact#personalTitle');
	my $firstname = firstOf($S, 'urn:ietf:rfc:2426#given-name',
		'http://www.w3.org/2006/vcard/ns#given-name',
		'http://www.w3.org/2006/03/hcard#given-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#given-name',
		'http://xmlns.com/foaf/0.1/givenname',
		'http://www.w3.org/2000/10/swap/pim/contact#firstName');
	my $surname = firstOf($S, 'urn:ietf:rfc:2426#family-name',
		'http://www.w3.org/2006/vcard/ns#family-name',
		'http://www.w3.org/2006/03/hcard#family-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#family-name',
		'http://xmlns.com/foaf/0.1/family_name',
		'http://xmlns.com/foaf/0.1/surname',
		'http://www.w3.org/2000/10/swap/pim/contact#lastName');
	my $honorific_sfx = firstOf($S, 'urn:ietf:rfc:2426#honorific-suffix',
		'http://www.w3.org/2006/vcard/ns#honorific-suffix',
		'http://www.w3.org/2006/03/hcard#honorific-suffix',
		'http://www.w3.org/2001/vcard-rdf/3.0#honorific-suffix',
		'http://www.w3.org/2000/10/swap/pim/contact#personalSuffix');

	$fn = "$honorifix_pfx $firstname $surname $honorific_sfx";
	$fn =~ s/(^\s+|\s+$)//g;
	$fn =~ s/\s+/ /g;
	if (length $fn)
		{ return $fn; }
		
	# Nickname
	return firstOf($S, 'urn:ietf:rfc:2426#nickname',
		'http://www.w3.org/2006/vcard/ns#nickname',
		'http://www.w3.org/2006/03/hcard#nickname',
		'http://www.w3.org/2001/vcard-rdf/3.0#nickname',
		'http://xmlns.com/foaf/0.1/nick',
		'http://www.w3.org/2000/10/swap/pim/contact#knownAs');
}

# Implements the atom "Person" construct, though actually we may be an
# organisation or other type of contact.
sub _AtomPerson
{
	my $triples = shift;
	my $subject = shift;
	my $atomtag = shift;
	my $indent;
	my $S = $triples->{$subject};

	$atomtag =~ /^(\s*)<(.*)>$/;
	$indent  = $1;
	$atomtag = $2;

	my $name  = outName($triples, $subject);
	my $email = firstOf($S,
		'urn:ietf:rfc:2426#email',
		'http://www.w3.org/2006/vcard/ns#email',
		'http://www.w3.org/2006/vcard/ns#workEmail',
		'http://www.w3.org/2006/vcard/ns#mobileEmail',
		'http://www.w3.org/2006/vcard/ns#unlabelledEmail',
		'http://www.w3.org/2006/03/hcard#email',
		'http://www.w3.org/2001/vcard-rdf/3.0#email',
		'http://www.w3.org/2000/10/swap/pim/contact#emailAddress',
		'http://xmlns.com/foaf/0.1/mbox',
		'http://xmlns.com/foaf/0.1/mbox_sha1sum');
	my $url   = firstOf($S, 'urn:ietf:rfc:2426#url',
		'http://www.w3.org/2006/vcard/ns#url',
		'http://www.w3.org/2006/03/hcard#url',
		'http://www.w3.org/2001/vcard-rdf/3.0#url',
		'http://xmlns.com/foaf/0.1/homepage',
		'http://xmlns.com/foaf/0.1/weblog',
		'http://xmlns.com/foaf/0.1/workplaceHomepage',
		'http://xmlns.com/foaf/0.1/workInfoHomepage',
		'http://xmlns.com/foaf/0.1/schoolHomepage',
		'http://xmlns.com/foaf/0.1/tipjar',
		'http://www.w3.org/2000/10/swap/pim/contact#homepage',
		'http://www.w3.org/2000/10/swap/pim/contact#publicHomepage');
	
	$email =~ s/(^mailto:|\?.*$)//g;
	
	my $rv = "$indent<$atomtag\n$indent    xmlns:hcard=\"urn:ietf:rfc:2426#\"\n$indent    xmlns:foaf=\"http://xmlns.com/foaf/0.1/\">\n";
	$rv .=   "$indent\t<name>".encode_entities_numeric($name)."</name>\n" if ($name);
	$rv .=   "$indent\t<email>".encode_entities_numeric($email)."</email>\n" if ($email);
	$rv .=   "$indent\t<uri>".encode_entities_numeric($url)."</uri>\n" if ($url);
	foreach my $p ($S->properties)
	{
		next if ($p eq 'urn:ietf:rfc:2426#n'
			|| $p eq 'urn:ietf:rfc:2426#fn'
			|| $p eq 'urn:ietf:rfc:2426#url'
			|| $p eq 'urn:ietf:rfc:2426#email'); # redundant
	
		my $P = $S->property($p);
		if ($P->namespace eq 'urn:ietf:rfc:2426#'
		||  $P->namespace eq 'http://xmlns.com/foaf/0.1/')
		{
			my $ns = 'hcard';
			$ns = 'foaf' if ($P->namespace eq 'http://xmlns.com/foaf/0.1/');
			
			foreach my $V (@{ $P->valObjects })
			{
				my $v = $V->toString;
				$rv .= "$indent\t<$ns:".$P->localname.">".encode_entities_numeric($v)
					."</$ns:".$P->localname.">\n";
			}
		}
	}
	$rv .= "$indent</$atomtag>\n";
	return $rv;
}

1;
