#!/usr/bin/perl

package Cognition::Export::FBInfo;
use Cognition::Export;
use Cognition::Export::Contact;
use MIME::Base64;
use URI;

sub is
{
	my $subject = shift;
	return unless ($subject);
	return
		unless (defined $subject->property('http://www.w3.org/1999/02/22-rdf-syntax-ns#type'));
	foreach my $Type ( @{$subject
		->property('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')
		->valObjects} )
	{
		my $type = $Type->toString;
		return 1
			if (($type eq 'http://www.w3.org/2002/12/cal/ical#Vfreebusy')
			||  ($type eq 'urn:ietf:rfc:2445#vfreebusy'));
	}
	return;
}

sub outputs
{
	return [
		{
			code  => 'ICS',
			name  => 'iCalendar',
			type  => 'text/directory',
			multi => 1
		}
	];
}

sub outICS
{
	return "BEGIN:VCALENDAR\r\n"
		. _outICS(@_)
		. "END:VCALENDAR\r\n";
}

sub multiICS
{
	my $triples = shift;
	my $rv = '';
	foreach my $s (keys %$triples)
	{
		$rv .= _outICS($triples, $s)
			if (defined $triples->{$s} && is($triples->{$s}));
	}

	return "BEGIN:VCALENDAR\r\n"
		. $rv
		. "END:VCALENDAR\r\n";
}

sub _outICS
{
	my $triples = shift;
	my $subject = shift;
	my $S = $triples->{$subject};
	my @rv = (['BEGIN','VFREEBUSY']);

	# UID
	my $uid = firstOf($S, 'urn:ietf:rfc:2445#uid',
		'http://www.w3.org/2002/12/cal/ical#uid',
		'http://purl.org/dc/terms/identifier',
		'http://purl.org/dc/elements/1.1/identifier');
	push @rv, ['UID', $uid]
		if ($uid);
	
	# Summary
	my $summary = firstOf($S, 'urn:ietf:rfc:2445#summary',
		'http://www.w3.org/2002/12/cal/ical#summary',
		'http://purl.org/dc/terms/title',
		'http://purl.org/dc/elements/1.1/title');
	push @rv, ['SUMMARY', $summary]
		if ($summary);
	
	# Freebusy
	my @fb = allOf($S, 'urn:ietf:rfc:2445#freebusy',
		'http://www.w3.org/2002/12/cal/ical#freebusy');
	foreach my $fb (@fb)
		{ push @rv, ['FREEBUSY', $fb]; }
	
	# Description
	my $desc = firstOf($S, 'urn:ietf:rfc:2445#comment',
		'http://www.w3.org/2002/12/cal/ical#comment',
		'http://purl.org/dc/terms/description',
		'http://purl.org/dc/elements/1.1/description',
		'http://purl.org/dc/terms/abstract');
	push @rv, ['COMMENT', $desc]
		if ($desc);
	
	# DTStart, DTEnd, DTStamp, Duration
	my $dt = firstOf($S, 'urn:ietf:rfc:2445#dtstart',
		'http://www.w3.org/2002/12/cal/ical#dtstart');
	push @rv, ['DTSTART', $dt] if ($dt);
	$dt = firstOf($S, 'urn:ietf:rfc:2445#dtend',
		'http://www.w3.org/2002/12/cal/ical#dtend');
	push @rv, ['DTEND', $dt] if ($dt);
	$dt = firstOf($S, 'urn:ietf:rfc:2445#dtstamp',
		'http://www.w3.org/2002/12/cal/ical#dtstamp');
	push @rv, ['DTSTAMP', $dt] if ($dt);
	$dt = firstOf($S, 'urn:ietf:rfc:2445#duration',
		'http://www.w3.org/2002/12/cal/ical#duration');
	push @rv, ['DURATION', $dt] if ($dt);
	
	# URL
	my $url = firstOf($S, 'urn:ietf:rfc:2445#url',
		'http://www.w3.org/2002/12/cal/ical#url');
	push @rv, ['URL', $url]
		if ($url);
	
	# Attendee
	my @agents = allOfObj($S, 'urn:ietf:rfc:2445#attendee',
		'http://www.w3.org/2002/12/cal/ical#attendee');
	foreach my $agent (@agents)
	{
		if ($agent->type && $triples->{$agent->toString})
		{
			my $agentSubject = $triples->{$agent->toString};
			if (Cognition::Export::Contact::is($agentSubject))
			{
				my $a = Cognition::Export::Contact::outVCF($triples, $agent->toString);
				push @rv, ['ATTENDEE',$a];
			}
			else
			{
				my $a = firstOf($agentSubject,
					'urn:ietf:rfc:2426#fn',
					'http://www.w3.org/2006/vcard/ns#fn',
					'http://www.w3.org/2006/03/hcard#fn',
					'http://www.w3.org/2001/vcard-rdf/3.0#fn',
					'http://xmlns.com/foaf/0.1/fullName',
					'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
					'http://purl.org/dc/terms/title',
					'http://purl.org/dc/elements/1.1/title');
				push @rv, ['ATTENDEE;VALUE=text',$a] if ($a);
			}
		}
		else
		{
			my $a = $agent->toString;
			$a =~ s/(^\s+|\s+$)//g;
			$a =~ s/\s+/ /g;
			push @rv, ['ATTENDEE;VALUE=text',$a] if ($a);
		}
	}
		
	# contacts
	my @agents = allOfObj($S, 'urn:ietf:rfc:2445#contact',
		'http://www.w3.org/2002/12/cal/ical#contact');
	foreach my $agent (@agents)
	{
		if ($agent->type && $triples->{$agent->toString})
		{
			my $agentSubject = $triples->{$agent->toString};
			if (Cognition::Export::Contact::is($agentSubject))
			{
				my $a = Cognition::Export::Contact::outVCF($triples, $agent->toString);
				push @rv, ['CONTACT',$a];
			}
			else
			{
				my $a = firstOf($agentSubject,
					'urn:ietf:rfc:2426#fn',
					'http://www.w3.org/2006/vcard/ns#fn',
					'http://www.w3.org/2006/03/hcard#fn',
					'http://www.w3.org/2001/vcard-rdf/3.0#fn',
					'http://xmlns.com/foaf/0.1/fullName',
					'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
					'http://purl.org/dc/terms/title',
					'http://purl.org/dc/elements/1.1/title');
				push @rv, ['CONTACT;VALUE=text',$a] if ($a);
			}
		}
		else
		{
			my $a = $agent->toString;
			$a =~ s/(^\s+|\s+$)//g;
			$a =~ s/\s+/ /g;
			push @rv, ['CONTACT;VALUE=text',$a] if ($a);
		}
	}
		
	# orgs
	my @agents = allOfObj($S, 'urn:ietf:rfc:2445#organizer',
		'http://www.w3.org/2002/12/cal/ical#organizer');
	foreach my $agent (@agents)
	{
		if ($agent->type && $triples->{$agent->toString})
		{
			my $agentSubject = $triples->{$agent->toString};
			if (Cognition::Export::Contact::is($agentSubject))
			{
				my $a = Cognition::Export::Contact::outVCF($triples, $agent->toString);
				push @rv, ['ORGANIZER',$a];
			}
			else
			{
				my $a = firstOf($agentSubject,
					'urn:ietf:rfc:2426#fn',
					'http://www.w3.org/2006/vcard/ns#fn',
					'http://www.w3.org/2006/03/hcard#fn',
					'http://www.w3.org/2001/vcard-rdf/3.0#fn',
					'http://xmlns.com/foaf/0.1/fullName',
					'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
					'http://purl.org/dc/terms/title',
					'http://purl.org/dc/elements/1.1/title');
				push @rv, ['ORGANIZER;VALUE=text',$a] if ($a);
			}
		}
		else
		{
			my $a = $agent->toString;
			$a =~ s/(^\s+|\s+$)//g;
			$a =~ s/\s+/ /g;
			push @rv, ['ORGANIZER;VALUE=text',$a] if ($a);
		}
	}
		
	# Done.
	push @rv, ['END','VFREEBUSY'];
	
	# Convert to text/directory syntax.
	my $rv = '';
	foreach my $line (@rv)
	{
		my ($k, $v, $soft) = @{$line};
		
		if ($k =~ /VALUE.uri$/i && $v =~ /^data\:/)
		{
			my $uri = URI->new($v);
			$v = encode_base64($uri->data, '');
			$k =~ s/VALUE.uri$/ENCODING\=b/i;
		}
		
		$v =~ s/\\/\\\\/g;
		$v =~ s/\r?\n/\\n/g;
		unless ($soft)
		{
			$v =~ s/\;/\\\;/g;
			$v =~ s/\,/\\\,/g;
		}
		
		$rv .= "$k:$v\r\n";
	}
	return $rv;
}

1;
